@@ -60,6 +60,7 @@ xstype xs_type(const xs_val *data);
 int xs_size(const xs_val *data);
 int xs_is_null(const xs_val *data);
 int xs_cmp(const xs_val *v1, const xs_val *v2);
+const xs_val *xs_or(const xs_val *v1, const xs_val *v2);
 xs_val *xs_dup(const xs_val *data);
 xs_val *xs_expand(xs_val *data, int offset, int size);
 xs_val *xs_collapse(xs_val *data, int offset, int size);
@@ -80,9 +81,9 @@ xs_str *xs_replace_in(xs_str *str, const char *sfrom, const char *sto, int times
 #define xs_replace_n(str, sfrom, sto, times) xs_replace_in(xs_dup(str), sfrom, sto, times)
 xs_str *xs_fmt(const char *fmt, ...);
 int xs_str_in(const char *haystack, const char *needle);
-int xs_starts_and_ends(const char *prefix, const char *str, const char *suffix);
-#define xs_startswith(str, prefix) xs_starts_and_ends(prefix, str, NULL)
-#define xs_endswith(str, suffix) xs_starts_and_ends(NULL, str, suffix)
+int xs_between(const char *prefix, const char *str, const char *suffix);
+#define xs_startswith(str, prefix) xs_between(prefix, str, NULL)
+#define xs_endswith(str, suffix) xs_between(NULL, str, suffix)
 xs_str *xs_crop_i(xs_str *str, int start, int end);
 xs_str *xs_lstrip_chars_i(xs_str *str, const char *chars);
 xs_str *xs_rstrip_chars_i(xs_str *str, const char *chars);
@@ -119,11 +120,16 @@ xs_dict *xs_dict_new(void);
 xs_dict *xs_dict_append(xs_dict *dict, const xs_str *key, const xs_val *value);
 xs_dict *xs_dict_prepend(xs_dict *dict, const xs_str *key, const xs_val *value);
 int xs_dict_next(const xs_dict *dict, const xs_str **key, const xs_val **value, int *ctxt);
-const xs_val *xs_dict_get_def(const xs_dict *dict, const xs_str *key, const xs_val *def);
-#define xs_dict_get(dict, key) xs_dict_get_def(dict, key, NULL)
+const xs_val *xs_dict_get(const xs_dict *dict, const xs_str *key);
+#define xs_dict_get_def(dict, key, def) xs_or(xs_dict_get(dict, key), def)
 xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key);
 xs_dict *xs_dict_set(xs_dict *dict, const xs_str *key, const xs_val *data);
-xs_dict *xs_dict_gc(xs_dict *dict);
+xs_dict *xs_dict_gc(const xs_dict *dict);
+
+const xs_val *xs_dict_get_path_sep(const xs_dict *dict, const char *path, const char *sep);
+#define xs_dict_get_path(dict, path) xs_dict_get_path_sep(dict, path, ".")
+xs_dict *xs_dict_set_path_sep(xs_dict *dict, const char *path, const xs_val *value, const char *sep);
+#define xs_dict_set_path(dict, path, value) xs_dict_set_path_sep(dict, path, value, ".")
 
 xs_val *xs_val_new(xstype t);
 xs_number *xs_number_new(double f);
@@ -151,6 +157,11 @@ unsigned int xs_hash_func(const char *data, int size);
 
 #define xs_is_true(v) (xs_type((v)) == XSTYPE_TRUE)
 #define xs_is_false(v) (xs_type((v)) == XSTYPE_FALSE)
+#define xs_not(v) xs_stock(xs_is_true((v)) ? XSTYPE_FALSE : XSTYPE_TRUE)
+
+#define xs_list_foreach(l, v) for (int ct_##__LINE__ = 0; xs_list_next(l, &v, &ct_##__LINE__); )
+#define xs_dict_foreach(l, k, v) for (int ct_##__LINE__ = 0; xs_dict_next(l, &k, &v, &ct_##__LINE__); )
+
 
 #ifdef XS_IMPLEMENTATION
 
@@ -347,6 +358,9 @@ int xs_is_null(const xs_val *data)
 int xs_cmp(const xs_val *v1, const xs_val *v2)
 /* compares two values */
 {
+    if (xs_type(v1) == XSTYPE_STRING && xs_type(v2) == XSTYPE_STRING)
+        return strcmp(v1, v2);
+
     int s1 = xs_size(v1);
     int s2 = xs_size(v2);
     int d = s1 - s2;
@@ -355,6 +369,13 @@ int xs_cmp(const xs_val *v1, const xs_val *v2)
 }
 
 
+const xs_val *xs_or(const xs_val *v1, const xs_val *v2)
+/* returns v1 if it's not NULL, else v2 */
+{
+    return v1 == NULL ? v2 : v1;
+}
+
+
 xs_val *xs_dup(const xs_val *data)
 /* creates a duplicate of data */
 {
@@ -581,7 +602,7 @@ int xs_str_in(const char *haystack, const char *needle)
 }
 
 
-int xs_starts_and_ends(const char *prefix, const char *str, const char *suffix)
+int xs_between(const char *prefix, const char *str, const char *suffix)
 /* returns true if str starts with prefix and ends with suffix */
 {
     int sz = strlen(str);
@@ -800,10 +821,10 @@ int xs_list_len(const xs_list *list)
 {
     XS_ASSERT_TYPE_NULL(list, XSTYPE_LIST);
 
-    int c = 0, ct = 0;
+    int c = 0;
     const xs_val *v;
 
-    while (xs_list_next(list, &v, &ct))
+    xs_list_foreach(list, v)
         c++;
 
     return c;
@@ -818,10 +839,10 @@ const xs_val *xs_list_get(const xs_list *list, int num)
     if (num < 0)
         num = xs_list_len(list) + num;
 
-    int c = 0, ct = 0;
+    int c = 0;
     const xs_val *v;
 
-    while (xs_list_next(list, &v, &ct)) {
+    xs_list_foreach(list, v) {
         if (c == num)
             return v;
 
@@ -909,11 +930,10 @@ int xs_list_in(const xs_list *list, const xs_val *val)
     XS_ASSERT_TYPE_NULL(list, XSTYPE_LIST);
 
     int n = 0;
-    int ct = 0;
     const xs_val *v;
     int sz = xs_size(val);
 
-    while (xs_list_next(list, &v, &ct)) {
+    xs_list_foreach(list, v) {
         if (sz == xs_size(v) && memcmp(val, v, sz) == 0)
             return n;
 
@@ -932,11 +952,10 @@ xs_str *xs_join(const xs_list *list, const char *sep)
     xs_str *s = NULL;
     const xs_val *v;
     int c = 0;
-    int ct = 0;
     int offset = 0;
     int ssz = strlen(sep);
 
-    while (xs_list_next(list, &v, &ct)) {
+    xs_list_foreach(list, v) {
         /* refuse to join non-string values */
         if (xs_type(v) == XSTYPE_STRING) {
             int sz;
@@ -1206,8 +1225,8 @@ xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key)
 }
 
 
-const xs_val *xs_dict_get_def(const xs_dict *dict, const xs_str *key, const xs_str *def)
-/* gets a value by key, or returns def */
+const xs_val *xs_dict_get(const xs_dict *dict, const xs_str *key)
+/* gets a value by key, or NULL */
 {
     if (xs_type(dict) == XSTYPE_DICT) {
         int *o = _xs_dict_locate(dict, key);
@@ -1221,7 +1240,7 @@ const xs_val *xs_dict_get_def(const xs_dict *dict, const xs_str *key, const xs_s
         }
     }
 
-    return def;
+    return NULL;
 }
 
 
@@ -1258,24 +1277,79 @@ int xs_dict_next(const xs_dict *dict, const xs_str **key, const xs_val **value,
 }
 
 
-xs_dict *xs_dict_gc(xs_dict *dict)
-/* collects garbage (leaked values) inside a dict */
+xs_dict *xs_dict_gc(const xs_dict *dict)
+/* creates a copy of dict, but garbage-collected */
 {
     xs_dict *nd = xs_dict_new();
     const xs_str *k;
     const xs_val *v;
-    int c = 0;
 
-    /* shamelessly create a new dict with the same content */
-    while (xs_dict_next(dict, &k, &v, &c))
-        nd = xs_dict_set(nd, k, v);
-
-    xs_free(dict);
+    xs_dict_foreach(dict, k, v) {
+        if (xs_type(v) == XSTYPE_DICT) {
+            xs *sd = xs_dict_gc(v);
+            nd = xs_dict_set(nd, k, sd);
+        }
+        else
+            nd = xs_dict_set(nd, k, v);
+    }
 
     return nd;
 }
 
 
+const xs_val *xs_dict_get_path_sep(const xs_dict *dict, const char *path, const char *sep)
+/* gets a value from dict given a path separated by sep */
+{
+    /* split by the separator */
+    xs *l = xs_split_n(path, sep, 1);
+
+    /* only one part? just get */
+    if (xs_list_len(l) == 1)
+        return xs_dict_get(dict, path);
+
+    const char *prefix = xs_list_get(l, 0);
+    const char *rest   = xs_list_get(l, 1);
+    const xs_dict *sd  = xs_dict_get(dict, prefix);
+
+    if (xs_type(sd) == XSTYPE_DICT)
+        return xs_dict_get_path_sep(sd, rest, sep);
+
+    return NULL;
+}
+
+
+xs_dict *xs_dict_set_path_sep(xs_dict *dict, const char *path, const xs_val *value, const char *sep)
+/* sets a value into dict given a path separated by sep;
+   intermediate dicts are created if needed */
+{
+    /* split by the separator */
+    xs *l = xs_split_n(path, sep, 1);
+
+    /* only one part? just set */
+    if (xs_list_len(l) == 1)
+        return xs_dict_set(dict, path, value);
+
+    const char *prefix = xs_list_get(l, 0);
+    const char *rest   = xs_list_get(l, 1);
+
+    xs *nd = NULL;
+
+    /* does the first part of path exist? */
+    const xs_dict *cd = xs_dict_get(dict, prefix);
+
+    if (xs_type(cd) == XSTYPE_DICT)
+        nd = xs_dup(cd);
+    else
+        nd = xs_dict_new();
+
+    /* move down the path */
+    nd = xs_dict_set_path_sep(nd, rest, value, sep);
+
+    /* set */
+    return xs_dict_set(dict, prefix, nd);
+}
+
+
 /** other values **/
 
 xs_val *xs_val_new(xstype t)
